package main

import (
	"context"
	"errors"
	"fmt"
	"io"
	"log"
	"net/http"
	"os"
	"os/signal"
	"sync"
	"syscall"
	"time"
)

const (
	shutdownTimeoutDuration = time.Second * 15
)

func main() {
	mainCtx, mainCtxCancel := context.WithCancel(context.Background())
	defer mainCtxCancel()
	mainWg := sync.WaitGroup{}

	mux := http.NewServeMux()
	mux.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		w.Write([]byte("👋"))
	})
	mux.HandleFunc("/upload", func(w http.ResponseWriter, r *http.Request) {
		f, err := os.Create(fmt.Sprintf("%d", time.Now().UnixNano()))
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
		defer f.Close()

		_, err = io.Copy(f, r.Body)
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			return
		}

		_, err = w.Write([]byte(f.Name()))
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
	})

	server := &http.Server{
		Addr:    ":1337",
		Handler: mux,
	}

	fmt.Printf("listening for HTTP traffic on port %s\n", server.Addr)
	go func() {
		if err := server.ListenAndServe(); err != nil && !errors.Is(err, http.ErrServerClosed) {
			log.Fatal(err)
		}
	}()

	mainWg.Add(1)
	go func() {
		defer mainWg.Done()

		someBackgroundProcess(mainCtx)
	}()

	wait := make(chan os.Signal, 1)
	signal.Notify(wait, syscall.SIGINT, syscall.SIGTERM)

	fmt.Println("waiting for exit signal")

	sig := <-wait

	fmt.Printf("exit signal received: %s\n", sig.String())

	mainCtxCancel()

	shutdownCtx, shutdownCtxCancel := context.WithTimeout(context.Background(), shutdownTimeoutDuration)
	defer shutdownCtxCancel()

	fmt.Println("shutting down HTTP server")

	err := server.Shutdown(shutdownCtx)
	if err != nil {
		log.Fatal(err)
	}

	fmt.Println("HTTP server shutdown")
	fmt.Println("waiting for process cleanup")

	// TODO: there is an opportunity here to have WaitGroup.Wait() wrapped to return a channel in such a way that you could set a maximum time for go-routines to drain the waitgroup via a context timeout through select
	mainWg.Wait()

	fmt.Println("goodbye")
}

func someBackgroundProcess(ctx context.Context) {
	for {
		select {
		case <-ctx.Done():
			return

		case <-time.After(time.Second * 10):
			fmt.Println("do something")
			time.Sleep(time.Second * 3) // simulate some work
		}
	}
}
